﻿//////////////////////////////////////////////
// Texture.h
//
//////////////////////////////////////////////

/// Defines / Macros -------------------------

#pragma once

/// Forward decl -----------------------------

namespace nkGraphics
{
	struct MapRequestDescriptor ;
	struct ResourceCpuDataDescriptor ;
	struct ResourceGpuCopyDescriptor ;
	struct UnmapRequestDescriptor ;
}

/// Includes ---------------------------------

// nkGraphics
#include "../Dll/DllDefines.h"

#include "../Buffers/MapResult.h"

#include "../Resources/ShaderResource.h"

#include "../Utilities/DxDefinesWrapper.h"

// nkExport
#include <NilkinsExport/Exportable/Exportable.h>

/// Internals --------------------------------

namespace nkGraphics
{
	enum class TEX_TYPE
	{
		TEX_2D = 0,
		TEX_3D,
		TEX_CUBEMAP
	} ;

	enum class TEX_RENDER_FLAG
	{
		NO_RENDER = 0,
		RENDER_TARGET,
		DEPTH_RENDER_TARGET
	} ;
}

/// Class ------------------------------------
	
namespace nkGraphics
{
	class DLL_GRAPHICS_EXPORT Texture : public ShaderResource, public nkExport::Exportable
	{
		protected :
		
			// Attributs
			// Données CPU si jamais
			ResourceCpuDataDescriptor* _cpuData ;

			// Le type réel interne
			TEX_TYPE _texType ;
			// Flag de rendu
			TEX_RENDER_FLAG _renderFlag ;

			// Témoin, savoir si nous sommes originaires d'un fichier
			bool _fromFile ;
			// Si on fait la correction Gamma
			// Cela sert pour les textures de diffuse, albedo, ou autre
			// Les textures codant des informations mathématiques (ou déjà en espace linéaire) ne doivent pas être corrigées
			bool _gammaCorrect ;
			// Savoir si nous sommes issus d'un back buffer quelconque nous empechant de faire les choses
			bool _backBufferTex ;
		
		public :
		
			// Constructeur, destructeur
			Texture () ;
			virtual ~Texture () ;

			// Surcharge
			virtual void setResourcePath (const nkMemory::StringView& path) override ;

			// Getters
			ResourceCpuDataDescriptor* getCpuData () const ;
			bool isFromFile () const ;
			bool isGammaCorrected () const ;
			bool isBackBufferTex () const ;
			TEX_TYPE getTexType () const ;
			TEX_RENDER_FLAG getRenderFlag () const ;
			unsigned int getNeededMipsForSize () const ;

			// Getters de paramètres
			virtual int getWidth () const = 0 ;
			virtual int getHeight () const = 0 ;
			virtual int getDepthOrArraySize () const = 0 ;
			virtual unsigned int getMipLevels () const = 0 ;
			virtual FORMAT getTextureFormat () const = 0 ;
			virtual CPU_ACCESS_FLAG getCpuAccessRights () const = 0 ;
			virtual USAGE getTextureUsage () const = 0 ;
			virtual BIND_FLAG getTextureBinding () const = 0 ;
			virtual RESOURCE_MISC_FLAG getMiscFlag () const = 0 ;
			virtual unsigned int getSampleCount () const = 0 ;
			virtual unsigned int getSampleQuality () const = 0 ;
			virtual bool getAutoMipGeneration () const = 0 ;

			// Setters
			void setCpuData (const ResourceCpuDataDescriptor& data) ;
			void setBackBufferTex (bool value) ;
			virtual void freeCpuData (bool freeDataPtr) ;
			virtual void setGammaCorrected (bool value) ;

			// Setters de paramètres
			virtual void setWidth (unsigned int width) = 0 ;
			virtual void setHeight (unsigned int height) = 0 ;
			virtual void setDepth (unsigned int depth) = 0 ;
			virtual void setArraySize (unsigned int size) = 0 ;
			virtual void setMipLevels (unsigned int value) = 0 ;
			virtual void setTextureFormat (FORMAT format) = 0 ;
			virtual void setCpuAccessRights (CPU_ACCESS_FLAG rights) = 0 ;
			virtual void setTextureUsage (USAGE usage) = 0 ;
			virtual void setTextureBinding (BIND_FLAG binding) = 0 ;
			virtual void setAutoMipGeneration (bool value) = 0 ;
			virtual void copyParams (const Texture& other) = 0 ;

			void setTextureFormatFromString (const nkMemory::StringView& str) ;

			// Préparations
			virtual void prepareAsTexture2d () ;
			virtual void prepareAsCubemap () ;
			virtual void prepareAsTexture3d () ;
			virtual void setRenderFlag (TEX_RENDER_FLAG value) ;

			virtual bool preLoadFromFile () = 0 ;

			// Lecture ou autres
			virtual MapResult map (const MapRequestDescriptor& requestDesc) = 0 ;
			virtual void unmap (const UnmapRequestDescriptor& requestDesc) = 0 ;
			virtual void gpuCopyFrom (Texture& other, const ResourceGpuCopyDescriptor& descriptor) = 0 ;

			// Les fonctions d'Import / Export
			virtual void exportClassToTree (nkExport::Node* rootNode) override ;
			virtual void importClassFromTree (nkExport::Node* rootNode) override ;
	} ;
}